from astropy.units import (
    Quantity,
    W,  # pyright: ignore[reportAttributeAccessIssue]
    mW,  # pyright: ignore[reportAttributeAccessIssue]
    Unit,
)
from astropy.units.physical import time
from gwpy.time import Time

from ..component import Component
from ..lockingstate import LockingState
from ..measurement import Measurement
from ..scatterer import Scatterer
from ..utils import download_dicttimeseries, download_timeseries

from ..metadata.measurement import Measurement as MetadataMeasurement
from ..metadata.component import Component as MetadataComponent
from ..metadata.scatterer import Scatterer as MetadataScatterer

from ..factory.component import ComponentFactory
from ..factory.scatterer import ScattererFactory
from ..factory.initializer import (
    InitializerScatterer,
)


class MeasurementFactory:
    """
    Factory class to create Measurement
    """

    @staticmethod
    def fetch(
        name: str,
        description: str,
        start_time: Time,
        duration: Quantity[time],
        initializer_metadata: list[InitializerScatterer],
    ) -> Measurement:
        """
        Fetch a measurement from online data

        Construct a Measurement object, containing all the data of this
        measurement
        Parameters
        ----------
        name : str
            Measurement name
        description : str
            Detailed description of the created measurement
        start_time : gwpy.Time
            Starting time of the created measurement
        duration : astropy.units.Quantity[time]
            Duration of the created measurement
        initializer_metadata : list[InitializerScatterer]
            List of component's and scatterer's pre-initialized
            metadata that will be used in the created measurement
        Returns
        -------
        Measurement : constructed Measurement from these metadata
        """
        channels_map: dict[str, Unit | None] = {
            "V1:META_ITF_LOCK_index": None,
            "V1:INJ_IMC_TRA_DC_mean": W,
            "V1:DET_B1p_DC_mean": mW,
        }
        data = download_dicttimeseries(
            channels=list(channels_map.keys()),
            start_time=start_time,
            duration=duration,
            units=list(channels_map.values()),
            frame="trend",
        )
        if sum(data["V1:META_ITF_LOCK_index"].diff().abs()) != 0:
            raise Exception(
                "The state of the interferometer change in this measurement"
            )
        state = LockingState(data["V1:META_ITF_LOCK_index"].data[0])
        injection_power = data["V1:INJ_IMC_TRA_DC_mean"].mean()
        dark_fringe_power = data["V1:DET_B1p_DC_mean"].mean()

        metadata = MetadataMeasurement(
            name=name,
            description=description,
            state=state,
            dark_fringe_power=dark_fringe_power,
            injection_power=injection_power,
            start_time=start_time,
            duration=duration,
        )

        components: list[Component] = []
        scatterers: list[Scatterer] = []
        for initializer_scatterer in initializer_metadata:
            component: Component | None = None
            if initializer_scatterer.associated_component is not None:
                component = ComponentFactory.fetch(
                    start_time=start_time,
                    duration=duration,
                    metadata=MetadataComponent(
                        name=initializer_scatterer.associated_component.name,
                        calibration_factor=initializer_scatterer.associated_component.calibration_factor,
                    ),
                )
                components.append(component)
            scatterers.append(
                ScattererFactory.fetch(
                    start_time=start_time,
                    duration=duration,
                    metadata=MetadataScatterer(
                        name=initializer_scatterer.name,
                        calibration_factor=initializer_scatterer.calibration_factor,
                        associated_component=component,
                        injection=initializer_scatterer.injection,
                    ),
                )
            )
        sensitivity = download_timeseries(
            channel="V1:Hrec_hoft_20000Hz",
            start_time=start_time,
            duration=duration,
            unit=None,
        )
        return Measurement(
            metadata=metadata,
            scatterers=scatterers,
            components=components,
            sensitivity=sensitivity,
        )
