from typing import final, override
from astropy.units import Unit
from gwpy.timeseries import TimeSeries
from h5py import Dataset, File, Group

from .section import Section
from .metadata.measurement import Measurement as Metadata
from .component import Component
from .scatterer import Scatterer
from .utils import dataset_to_timeseries


@final
class Measurement(Section):
    """
    A Measurement contains data from a given starting time and
    duration. It allows to regroup the sensitivity, a set of scatterers
    and components movements, and the state of the detector
    (specifically the injection power, the dark fringe offset and the
    lock index)
    """

    def __init__(
        self,
        metadata: Metadata,
        scatterers: list[Scatterer],
        components: list[Component],
        sensitivity: TimeSeries,
    ):
        """
        Parameters
        ----------
        metadata : vda.package.metadata.Measurement
            Metadata associated to this data
        scatterers : list[vda.package.Scatterer]
            List of scatterer in this measurement
        components : list[vda.package.Component]
            List of components in this measurement
        sensitivity : TimeSeries[unitless]
            Hrec signal in the temporal space
        Warnings
        --------
        It's an internal class, of whose instances are not produced via
        a direct call, but by using helper functions
        See Also
        --------
        bed.MetadataMeasurement : associated metadata type
        """
        self.metadata = metadata
        self.scatterers = scatterers
        self.components = components
        self.sensitivity = sensitivity

    @override
    @staticmethod
    def from_hdf5(store: File | Group) -> "Measurement":
        scatterers_group = store.get("scatterers")
        components_group = store.get("components")
        sensitivity_dataset = store.get("sensitivity")

        error = "{} should exists, and not be {}"
        if not isinstance(scatterers_group, Group):
            raise ValueError(
                error.format("scatterers", type(scatterers_group))
            )
        if not isinstance(components_group, Group):
            raise ValueError(
                error.format("components", type(components_group))
            )
        if not isinstance(sensitivity_dataset, Dataset):
            raise ValueError(
                error.format("senstivity", type(sensitivity_dataset))
            )
        scatterers = scatterers_group.values()
        components = components_group.values()
        sensitivity = dataset_to_timeseries(sensitivity_dataset)

        components_ = [
            Component.from_hdf5(component) for component in components
        ]
        return Measurement(
            Metadata.from_hdf5(store),
            [
                Scatterer.from_hdf5(
                    scatterer, components_
                )  # scatterer movement can be defined in respect to a component movement
                for scatterer in scatterers
            ],
            components_,
            sensitivity,
        )

    @override
    def to_hdf5(self, store: File | Group) -> Group | File:
        _ = self.metadata.to_hdf5(store)
        for component in self.components:
            _ = component.to_hdf5(
                store.require_group("components").create_group(
                    component.metadata.name
                )
            )
        for scatterer in self.scatterers:
            _ = scatterer.to_hdf5(
                store.require_group("scatterers").create_group(
                    scatterer.metadata.name
                )
            )
        dataset_store = store.create_dataset(
            name="sensitivity", data=self.sensitivity
        )
        if type(self.sensitivity.unit) is Unit:
            dataset_store.attrs["unit"] = (
                self.sensitivity.unit.to_string()
            )
        return store
