from typing import Literal
from astropy.units import Quantity, Unit, s, Hz  # pyright: ignore[reportAttributeAccessIssue]
from astropy.units.physical import time
from h5py import Dataset
from numpy import diff, mean
from requests import HTTPError

from gwpy.time import Time, to_gps
from gwpy.timeseries import TimeSeries, TimeSeriesDict

try:
    from virgotools.frame_lib import FrameFile, FrVect  # pyright: ignore[reportMissingImports]

    virgotool = True
except ImportError:
    virgotool = False
    pass


def dataset_to_timeseries(dataset: Dataset) -> TimeSeries:
    """
    Convert a h5df dataset to a TimeSeries keeping the units right
    Parameters
    ----------
    dataset : Dataset
        Dataset to convert
    Returns
    -------
    TimeSeries : TimeSeries with a good unit
    """
    unit = None
    if "units" in dataset.attrs.keys():
        units = dataset.attrs["units"]
        if isinstance(units, str):
            unit = Unit(units)
        else:
            raise ValueError(
                "cannot convert {} to a unit".format(units)
            )
    return TimeSeries(
        data=dataset[1],
        unit=unit,
        t0=dataset[0, 0],
        dt=mean(diff(dataset[0])),
    )


def download_timeseries(
    channel: str,
    start_time: Time,
    duration: Quantity[time],
    unit: Unit | None = None,
    frame: Literal["raw"] | Literal["trend"] = "raw",
) -> TimeSeries:
    """
    Download a TimeSeries from an online channel
    Parameters
    ----------
    channel : str
        Channel name
    start_time : Time
        Starting time
    duration : Quantity[time]
        Duration
    unit : Unit | None
        Unit of the TimeSeries (default ot None)
    frame : raw|trend
        Which frame to take the data from. Default to raw
    Notes
    -----
    It first try to use gwpy, and fallback to virgotools
    """
    try:
        return TimeSeries.get(
            channel=channel,
            start=to_gps(start_time),
            end=to_gps(start_time + duration),
        )
    except HTTPError:
        if virgotool:
            vect: FrVect | None = None  # pyright: ignore[reportPossiblyUnboundVariable]
            with FrameFile(frame) as ffl:  # pyright: ignore[reportPossiblyUnboundVariable]
                vect = ffl.getChannel(  # pyright: ignore[reportPossiblyUnboundVariable]
                    channel, to_gps(start_time), duration.to(s)
                )
            if vect is None:  # pyright: ignore[reportPossiblyUnboundVariable]
                raise RuntimeError(
                    "cannot get data vector from online data"
                )
            return TimeSeries(
                data=vect.data,  # pyright: ignore[reportPossiblyUnboundVariable]
                unit=unit,
                t0=start_time,
                sample_rate=vect.fsample * Hz,  # pyright: ignore[reportPossiblyUnboundVariable]
            )
        else:
            raise RuntimeError(
                "virgotools not available in this environment, cannot download online Virgo data"
            )


def download_dicttimeseries(
    channels: list[str],
    start_time: Time,
    duration: Quantity[time],
    units: list[Unit | None] | None = None,
    frame: Literal["raw"] | Literal["trend"] = "raw",
) -> TimeSeriesDict:
    """
    Download multiple TimeSeries at the same time to be more efficient
    Parameters
    ----------
    channels : list[str]
        List of channel name
    start_time : Time
        Starting time
    duration : Quantity[time]
        Duration
    units : list[Unit | None] | None
        Units of the different channels. If None, no unit for all
        channels. Default to None
    frame : raw|trend
        Which frame to take the data from. Default to raw
    Notes
    -----
    It first try to use gwpy, and fallback to virgotools
    """
    try:
        return TimeSeriesDict.get(
            channels=channels,
            start=to_gps(start_time),
            end=to_gps(start_time + duration),
        )
    except HTTPError:
        if virgotool:
            vects: list[FrVect] = []  # pyright: ignore[reportPossiblyUnboundVariable]
            with FrameFile(frame) as ffl:  # pyright: ignore[reportPossiblyUnboundVariable]
                for channel in channels:
                    vects.append(
                        ffl.getChannel(
                            channel, to_gps(start_time), duration.to(s)
                        )
                    )
            timeseries_dict: dict[str, TimeSeries] = dict()
            for i in range(len(vects)):
                unit = None
                if units is not None:
                    unit = units[i]
                timeseries_dict[channels[i]] = TimeSeries(
                    data=vects[i].data,
                    unit=unit,
                    t0=start_time,
                    sample_rate=vects[i].fsample * Hz,
                )
            return TimeSeriesDict(timeseries_dict)
        else:
            raise RuntimeError(
                "virgotools not available in this environment, cannot download online Virgo data"
            )
